@@ -1,7 +1,7 @@
 //
 //    This file is part of Dire Wolf, an amateur radio packet TNC.
 //
-//    Copyright (C) 2013  John Langner, WB2OSZ
+//    Copyright (C) 2013, 2015  John Langner, WB2OSZ
 //
 //    This program is free software: you can redistribute it and/or modify
 //    it under the terms of the GNU General Public License as published by
@@ -28,11 +28,20 @@
  * Description:	Establish connection with multiple servers and 
  *		compare results side by side.
  *
- * Usage:	aclients  8000=AGWPE  8002=DireWolf  COM1=D710A
+ * Usage:	aclients port1=name1 port2=name2 ...
+ *
+ * Example:	aclients  8000=AGWPE  192.168.1.64:8002=DireWolf  COM1=D710A
  *
  *		This will connect to multiple physical or virtual
  *		TNCs, read packets from them, and display results.
  *
+ *		Each port can have the following forms:
+ *
+ *		* host-name:tcp-port
+ *		* ip-addr:tcp-port
+ *		* tcp-port
+ *		* serial port name (e.g.  COM1, /dev/ttyS0)
+ *
  *---------------------------------------------------------------*/
 
 
@@ -40,7 +49,6 @@
 /*
  * Native Windows:	Use the Winsock interface.
  * Linux:		Use the BSD socket interface.
- * Cygwin:		Can use either one.
  */
 
 
@@ -59,6 +67,7 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <arpa/inet.h>
 #include <netinet/in.h>
 #include <netinet/tcp.h>
 #include <fcntl.h>
@@ -71,6 +80,7 @@
 #include <assert.h>
 #include <ctype.h>
 #include <string.h>
+#include <time.h>
 
 
 #include "direwolf.h"
@@ -115,7 +125,7 @@ static char * ia_to_text (int  Family, void * pAddr, char * pStringBuf, size_t S
 	  case AF_INET:
 	    sa4 = (struct sockaddr_in *)pAddr;
 #if __WIN32__
-	    sprintf (pStringBuf, "%d.%d.%d.%d", sa4->sin_addr.S_un.S_un_b.s_b1,
+	    snprintf (pStringBuf, StringBufSize, "%d.%d.%d.%d", sa4->sin_addr.S_un.S_un_b.s_b1,
 						sa4->sin_addr.S_un.S_un_b.s_b2,
 						sa4->sin_addr.S_un.S_un_b.s_b3,
 						sa4->sin_addr.S_un.S_un_b.s_b4);
@@ -126,7 +136,7 @@ static char * ia_to_text (int  Family, void * pAddr, char * pStringBuf, size_t S
 	  case AF_INET6:
 	    sa6 = (struct sockaddr_in6 *)pAddr;
 #if __WIN32__
-	    sprintf (pStringBuf, "%x:%x:%x:%x:%x:%x:%x:%x",  
+	    snprintf (pStringBuf, StringBufSize, "%x:%x:%x:%x:%x:%x:%x:%x",
 					ntohs(((unsigned short *)(&(sa6->sin6_addr)))[0]),
 					ntohs(((unsigned short *)(&(sa6->sin6_addr)))[1]),
 					ntohs(((unsigned short *)(&(sa6->sin6_addr)))[2]),
@@ -140,7 +150,7 @@ static char * ia_to_text (int  Family, void * pAddr, char * pStringBuf, size_t S
 #endif
 	    break;
 	  default:
-	    sprintf (pStringBuf, "Invalid address family!");
+	    snprintf (pStringBuf, StringBufSize, "Invalid address family!");
 	}
 	assert (strlen(pStringBuf) < StringBufSize);
 	return pStringBuf;
@@ -155,14 +165,7 @@ static char * ia_to_text (int  Family, void * pAddr, char * pStringBuf, size_t S
  * Purpose:   	Start up multiple client threads listening to different
  *		TNCs.   Print packets.  Tally up statistics.
  *
- * Usage:	aclients  8000=AGWPE  8002=DireWolf  COM1=D710A
- *
- *		Each command line argument is TCP port number or a 
- *		serial port name.  Follow by = and a text description
- *		of what is connected.
- *
- *		For now, everything is assumed to be on localhost.
- *		Maybe someday we might recognize host:port=description.
+ * Usage:	Described above.
  *
  *---------------------------------------------------------------*/
 
@@ -172,9 +175,17 @@ static char * ia_to_text (int  Family, void * pAddr, char * pStringBuf, size_t S
 
 static int num_clients;
 
-static char hostname[MAX_CLIENTS][50];
-static char port[MAX_CLIENTS][30];
-static char description[MAX_CLIENTS][50];
+static char hostname[MAX_CLIENTS][50];		/* DNS host name or IPv4 address. */
+						/* Some of the code is there for IPv6 but */
+						/* needs more work. */
+						/* Defaults to "localhost" if not specified. */
+
+static char port[MAX_CLIENTS][30];		/* If it begins with a digit, it is considered */
+						/* a TCP port number at the hostname.  */
+						/* Otherwise, we treat it as a serial port name. */
+
+static char description[MAX_CLIENTS][50];	/* Name used in the output. */
+
 
 #if __WIN32__
 	static HANDLE client_th[MAX_CLIENTS];
@@ -221,23 +232,41 @@ int main (int argc, char *argv[])
 	for (j=0; j<num_clients; j++) {
 	  char stemp[100];
 	  char *p;
-	
-	  strcpy (stemp, argv[j+1]);
+
+/* Each command line argument should be of the form "port=description." */
+
+	  strlcpy (stemp, argv[j+1], sizeof(stemp));
 	  p = strtok (stemp, "=");
 	  if (p == NULL) {
 	    printf ("Internal error 1\n");
 	    exit (1);
 	  }
-	  strcpy (hostname[j], "localhost");
-	  strcpy (port[j], p);
+	  strlcpy (hostname[j], "localhost", sizeof(hostname[j]));
+	  strlcpy (port[j], p, sizeof(port[j]));
 	  p = strtok (NULL, "=");
 	  if (p == NULL) {
 	    printf ("Missing description after %s\n", port[j]);
 	    exit (1);
 	  }
-	  strcpy (description[j], p);
+	  strlcpy (description[j], p, sizeof(description[j]));
+
+/* If the port contains ":" split it into hostname (or addr) and port number. */
+/* Haven't thought about IPv6 yet. */
+
+	  strlcpy (stemp, port[j], sizeof(stemp));
+
+	  char *h;
+
+	  h = strtok (stemp, ":");
+	  if (h != NULL) {
+	    p = strtok (NULL, ":");
+	    if (p != NULL) {
+	      strlcpy (hostname[j], h, sizeof(hostname[j]));
+	      strlcpy (port[j], p, sizeof(port[j]));
+	    }
+	  }
 	}
-	
+
 	//printf ("_WIN32_WINNT = %04x\n", _WIN32_WINNT);
 	//for (j=0; j<num_clients; j++) {
 	//  printf ("%s,%s,%s\n", hostname[j], port[j], description[j]);
@@ -252,6 +281,10 @@ int main (int argc, char *argv[])
 
 
 	for (j=0; j<num_clients; j++) {
+
+/* If port begins with digit, consider it to be TCP. */
+/* Otherwise, treat as serial port name. */
+
 #if __WIN32__
 	  if (isdigit(port[j][0])) {
 	    client_th[j] = (HANDLE)_beginthreadex (NULL, 0, client_thread_net, (void *)j, 0, NULL);
@@ -575,15 +608,18 @@ static void * client_thread_net (void *arg)
 	    char result[400];
 	    char *p;
 	    int col, len;
+	    alevel_t alevel;
 
 	    //printf ("server %d, portx = %d\n", my_index, mon_cmd.portx);
 
-	    use_chan == mon_cmd.portx;
-	    pp = ax25_from_frame ((unsigned char *)(data+1), mon_cmd.data_len-1, -1);
+	    use_chan = mon_cmd.portx;
+	    memset (&alevel, 0xff, sizeof(alevel));
+	    pp = ax25_from_frame ((unsigned char *)(data+1), mon_cmd.data_len-1, alevel);
+	    assert (pp != NULL);
 	    ax25_format_addrs (pp, result);
 	    info_len = ax25_get_info (pp, (unsigned char **)(&pinfo));
 	    pinfo[info_len] = '\0';
-	    strcat (result, pinfo);
+	    strlcat (result, pinfo, sizeof(result));
 	    for (p=result; *p!='\0'; p++) {
 	      if (! isprint(*p)) *p = ' ';
 	    }
