@@ -74,6 +74,11 @@ struct packet_s {
 
 	int magic1;		/* for error checking. */
 
+	int seq;		/* unique sequence number for debugging. */
+
+	double release_time;	/* Time stamp in format returned by dtime_now(). */
+				/* When to release from the SATgate mode delay queue. */
+
 #define MAGIC 0x41583235
 
 	struct packet_s *nextp;	/* Pointer to next in queue. */
@@ -144,13 +149,15 @@ typedef struct packet_s *packet_t;
 
 #ifdef AX25_PAD_C	/* Keep this hidden - implementation could change. */
 
+extern packet_t ax25_new (void);
+
 /*
  * APRS always has one control octet of 0x03 but the more
  * general AX.25 case is one or two control bytes depending on
  * "modulo 128 operation" is in effect.  Unfortunately, it seems
  * this can be determined only by examining the XID frames and 
  * keeping this information for each connection.
- * We can assume 1 for our purposes.
+ * We can assume 1 for our current purposes.
  */
 
 static inline int ax25_get_control_offset (packet_t this_p) 
@@ -161,10 +168,11 @@ static inline int ax25_get_control_offset (packet_t this_p)
 
 static inline int ax25_get_num_control (packet_t this_p)
 {
-	return (1);
+	return (1);	// TODO: always be 1 for U frame.  More complicated for I and S.
 }
 
 
+
 /*
  * APRS always has one protocol octet of 0xF0 meaning no level 3
  * protocol but the more general case is 0, 1 or 2 protocol ID octets.
@@ -196,14 +204,15 @@ static int ax25_get_num_pid (packet_t this_p)
 
 
 /*
- * APRS always has an Information field with at least one octet for the
- * Data Type Indicator.  AX.25 has this for only 5 frame types depending
- * on the control field.
+ * AX.25 has info field for 5 frame types depending on the control field.
+ *
  *	xxxx xxx0	I
- *	000x 0011	UI
+ *	000x 0011	UI		(which includes APRS)
  *	101x 1111	XID
  *	111x 0011	TEST
  *	100x 0111	FRMR
+ *
+ * APRS always has an Information field with at least one octet for the Data Type Indicator.  
  */
 
 static inline int ax25_get_info_offset (packet_t this_p) 
@@ -228,22 +237,91 @@ static int ax25_get_num_info (packet_t this_p)
 #endif
 
 
+typedef enum ax25_modulo_e { modulo_8 = 8, modulo_128 = 128 } ax25_modulo_t;
+
+typedef enum ax25_frame_type_e {
+
+	frame_type_I,		// Information
+
+	frame_type_S_RR,	// Receive Ready - System Ready To Receive
+	frame_type_S_RNR,	// Receive Not Ready - TNC Buffer Full
+	frame_type_S_REJ,	// Reject Frame - Out of Sequence or Duplicate
+	frame_type_S_SREJ,	// Selective Reject - Request single frame repeat
+
+	frame_type_U_SABME,	// Set Async Balanced Mode, Extended
+	frame_type_U_SABM,	// Set Async Balanced Mode
+	frame_type_U_DISC,	// Disconnect
+	frame_type_U_DM,	// Disconnect Mode
+	frame_type_U_UA,	// Unnumbered Acknowledge
+	frame_type_U_FRMR,	// Frame Reject
+	frame_type_U_UI,	// Unnumbered Information
+	frame_type_U_XID,	// Exchange Identification
+	frame_type_U_TEST,	// Test
+	frame_type_U,		// other Unnumbered, not used by AX.25.
+
+	frame_not_AX25		// Could not get control byte from frame.
+
+} ax25_frame_type_t;
+	
+
+/* 
+ * Originally this was a single number. 
+ * Let's try something new in version 1.2.
+ * Also collect AGC values from the mark and space filters.
+ */
+
+typedef struct alevel_s {
+
+	int rec;
+	int mark;
+	int space;
+	//float ms_ratio;	// TODO: take out after temporary investigation.
+} alevel_t;
 
 
+#ifndef AXTEST
+// TODO: remove this?
+#define AX25MEMDEBUG 1
+#endif
 
-//static packet_t ax25_new (void);
 
-extern void ax25_delete (packet_t pp);
+#if AX25MEMDEBUG	// to investigate a memory leak problem
 
-extern void ax25_clear (packet_t pp);
 
-extern packet_t ax25_from_text (char *, int strict);
+extern void ax25memdebug_set(void);
+extern int ax25memdebug_get (void);
+extern int ax25memdebug_seq (packet_t this_p);
 
-extern packet_t ax25_from_frame (unsigned char *data, int len, int alevel);
+
+extern packet_t ax25_from_text_debug (char *monitor, int strict, char *src_file, int src_line);
+#define ax25_from_text(m,s) ax25_from_text_debug(m,s,__FILE__,__LINE__)
+
+extern packet_t ax25_from_frame_debug (unsigned char *data, int len, alevel_t alevel, char *src_file, int src_line);
+#define ax25_from_frame(d,l,a) ax25_from_frame_debug(d,l,a,__FILE__,__LINE__);
+
+extern packet_t ax25_dup_debug (packet_t copy_from, char *src_file, int src_line);
+#define ax25_dup(p) ax25_dup_debug(p,__FILE__,__LINE__);
+
+extern void ax25_delete_debug (packet_t pp, char *src_file, int src_line);
+#define ax25_delete(p) ax25_delete_debug(p,__FILE__,__LINE__);
+
+#else
+
+extern packet_t ax25_from_text (char *monitor, int strict);
+
+extern packet_t ax25_from_frame (unsigned char *data, int len, alevel_t alevel);
 
 extern packet_t ax25_dup (packet_t copy_from);
 
-extern int ax25_parse_addr (char *in_addr, int strict, char *out_addr, int *out_ssid, int *out_heard);
+extern void ax25_delete (packet_t pp);
+
+#endif
+
+
+
+
+extern int ax25_parse_addr (int position, char *in_addr, int strict, char *out_addr, int *out_ssid, int *out_heard);
+extern int ax25_check_addresses (packet_t pp);
 
 extern packet_t ax25_unwrap_third_party (packet_t from_pp);
 
@@ -254,7 +332,8 @@ extern void ax25_remove_addr (packet_t this_p, int n);
 extern int ax25_get_num_addr (packet_t pp);
 extern int ax25_get_num_repeaters (packet_t this_p);
 
-extern void ax25_get_addr_with_ssid (packet_t pp, int n, char *);
+extern void ax25_get_addr_with_ssid (packet_t pp, int n, char *station);
+extern void ax25_get_addr_no_ssid (packet_t pp, int n, char *station);
 
 extern int ax25_get_ssid (packet_t pp, int n);
 extern void ax25_set_ssid (packet_t this_p, int n, int ssid);
@@ -275,15 +354,21 @@ extern int ax25_get_dti (packet_t this_p);
 
 extern packet_t ax25_get_nextp (packet_t this_p);
 
+extern void ax25_set_release_time (packet_t this_p, double release_time);
+extern double ax25_get_release_time (packet_t this_p);
+
 extern void ax25_format_addrs (packet_t pp, char *);
 
 extern int ax25_pack (packet_t pp, unsigned char result[AX25_MAX_PACKET_LEN]);
 
+extern ax25_frame_type_t ax25_frame_type (packet_t this_p, ax25_modulo_t modulo, char *desc, int *pf, int *nr, int *ns); 
+
 extern void ax25_hex_dump (packet_t this_p);
 
 extern int ax25_is_aprs (packet_t pp);
 
 extern int ax25_get_control (packet_t this_p); 
+extern int ax25_get_c2 (packet_t this_p); 
 
 extern int ax25_get_pid (packet_t this_p);
 
@@ -293,6 +378,9 @@ extern unsigned short ax25_m_m_crc (packet_t pp);
 
 extern void ax25_safe_print (char *, int, int ascii_only);
 
+#define AX25_ALEVEL_TO_TEXT_SIZE 32	// overkill but safe.
+extern int ax25_alevel_to_text (alevel_t alevel, char text[AX25_ALEVEL_TO_TEXT_SIZE]);
+
 
 #endif /* AX25_PAD_H */
 
