@@ -32,19 +32,27 @@ enum sendto_type_e { SENDTO_XMIT, SENDTO_IGATE, SENDTO_RECV };
 
 struct misc_config_s {
 
-	int num_channels;	/* Number of radio channels. */
-
-	int agwpe_port;		/* Port number for the AGW TCPIP Socket Interface */
-	int kiss_port;		/* Port number for the KISS protocol. */
+	int agwpe_port;		/* Port number for the "AGW TCPIP Socket Interface" */
+	int kiss_port;		/* Port number for the "KISS" protocol. */
 	int enable_kiss_pt;	/* Enable pseudo terminal for KISS. */
 				/* Want this to be off by default because it hangs */
 				/* after a while if nothing is reading from other end. */
 
-	char nullmodem[40];	/* Serial port name for our end of the */
+	char nullmodem[20];	/* Serial port name for our end of the */
 				/* virtual null modem for native Windows apps. */
 
-	char nmea_port[40];	/* Serial port name for NMEA communication with GPS */
-				/* receiver and/or mapping application. */
+	char gpsnmea_port[20];	/* Serial port name for reading NMEA sentences from GPS. */
+				/* e.g. COM22, /dev/ttyACM0 */
+
+	char gpsd_host[20];	/* Host for gpsd server. */
+				/* e.g. localhost, 192.168.1.2 */
+
+	int gpsd_port;		/* Port number for gpsd server. */
+				/* Default is  2947. */
+
+				/* e.g. COM22, /dev/ttyACM0 */
+	char nmea_port[20];	/* Serial port name for NMEA communication with GPS */
+				/* receiver and/or mapping application. Change this. */
 
 	char logdir[80];	/* Directory for saving activity logs. */
 
@@ -99,6 +107,9 @@ struct misc_config_s {
 	  char *custom_info;	/* Info part for handcrafted custom beacon. */
 				/* Ignore the rest below if this is set. */
 
+	  char *custom_infocmd;	/* Command to generate info part. */
+				/* Again, other options below are then ignored. */
+
 	  int messaging;	/* Set messaging attribute for position report. */
 				/* i.e. Data Type Indicator of '=' rather than '!' */
 
@@ -121,6 +132,7 @@ struct misc_config_s {
 	  float offset;		/* MHz. */
 	
 	  char *comment;	/* Comment or NULL. */
+	  char *commentcmd;	/* Command to append more to Comment or NULL. */
 
 
 	} beacon[MAX_BEACONS];
