@@ -1,7 +1,7 @@
 //
 //    This file is part of Dire Wolf, an amateur radio packet TNC.
 //
-//    Copyright (C) 2011-2014  John Langner, WB2OSZ
+//    Copyright (C) 2011-2014, 2015  John Langner, WB2OSZ
 //
 //    This program is free software: you can redistribute it and/or modify
 //    it under the terms of the GNU General Public License as published by
@@ -105,6 +105,11 @@
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
+#ifdef __OpenBSD__
+#include <errno.h>
+#else
+#include <sys/errno.h>
+#endif
 #endif
 
 #include <unistd.h>
@@ -133,8 +138,6 @@ static int client_sock;		/* File descriptor for socket for */
 				/* Set to -1 if not connected. */
 				/* (Don't use SOCKET type because it is unsigned.) */
 
-static int num_channels;	/* Number of radio ports. */
-
 
 static void * connect_listen_thread (void *arg);
 static void * kissnet_listen_thread (void *arg);
@@ -161,6 +164,8 @@ void kiss_net_set_debug (int n)
  *				  Main program has default of 8000 but allows
  *				  an alternative to be specified on the command line
  *
+ *				0 means disable.  New in version 1.2.
+ *
  * Outputs:	
  *
  * Description:	This starts two threads:
@@ -192,8 +197,13 @@ void kissnet_init (struct misc_config_s *mc)
 	memset (&kf, 0, sizeof(kf));
 	
 	client_sock = -1;
-	num_channels = mc->num_channels;
 
+	if (kiss_port == 0) {
+	  text_color_set(DW_COLOR_INFO);
+	  dw_printf ("Disabled KISS network client port.\n");
+	  return;
+	}
+	
 /*
  * This waits for a client to connect and sets client_sock.
  */
@@ -265,7 +275,7 @@ static void * connect_listen_thread (void *arg)
 	SOCKET listen_sock;  
 	WSADATA wsadata;
 
-	sprintf (kiss_port_str, "%d", (int)(long)arg);
+	snprintf (kiss_port_str, sizeof(kiss_port_str), "%d", (int)(long)arg);
 #if DEBUG
 	text_color_set(DW_COLOR_DEBUG);
         dw_printf ("DEBUG: kissnet port = %d = '%s'\n", (int)(long)arg, kiss_port_str);
@@ -315,7 +325,9 @@ static void * connect_listen_thread (void *arg)
 	err = bind( listen_sock, ai->ai_addr, (int)ai->ai_addrlen);
 	if (err == SOCKET_ERROR) {
 	  text_color_set(DW_COLOR_ERROR);
-          dw_printf("Bind failed with error: %d\n", WSAGetLastError());
+          dw_printf("Bind failed with error: %d\n", WSAGetLastError());		// TODO: provide corresponding text.
+	  dw_printf("Some other application is probably already using port %s.\n", kiss_port_str);
+	  dw_printf("Try using a different port number with KISSPORT in the configuration file.\n");
           freeaddrinfo(ai);
           closesocket(listen_sock);
           WSACleanup();
@@ -362,12 +374,15 @@ static void * connect_listen_thread (void *arg)
 
  	}
 
-#else
+
+#else		/* End of Windows case, now Linux. */
+
 
     	struct sockaddr_in sockaddr; /* Internet socket address stuct */
     	socklen_t sockaddr_size = sizeof(struct sockaddr_in);
 	int kiss_port = (int)(long)arg;
 	int listen_sock;  
+	int bcopt = 1;
 
 	listen_sock= socket(AF_INET,SOCK_STREAM,0);
 	if (listen_sock == -1) {
@@ -376,6 +391,14 @@ static void * connect_listen_thread (void *arg)
 	  return (NULL);
 	}
 
+	/* Version 1.3 - as suggested by G8BPQ. */
+	/* Without this, if you kill the application then try to run it */
+	/* again quickly the port number is unavailable for a while. */
+	/* Don't try doing the same thing On Windows; It has a different meaning. */
+	/* http://stackoverflow.com/questions/14388706/socket-options-so-reuseaddr-and-so-reuseport-how-do-they-differ-do-they-mean-t */
+
+        setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, (const char *)&bcopt, 4);
+
     	sockaddr.sin_addr.s_addr = INADDR_ANY;
     	sockaddr.sin_port = htons(kiss_port);
     	sockaddr.sin_family = AF_INET;
@@ -387,7 +410,10 @@ static void * connect_listen_thread (void *arg)
 
         if (bind(listen_sock,(struct sockaddr*)&sockaddr,sizeof(sockaddr))  == -1) {
 	  text_color_set(DW_COLOR_ERROR);
-    	  perror ("connect_listen_thread: Bind failed");
+          dw_printf("Bind failed with error: %d\n", errno);	
+          dw_printf("%s\n", strerror(errno));
+	  dw_printf("Some other application is probably already using port %d.\n", kiss_port);
+	  dw_printf("Try using a different port number with KISSPORT in the configuration file.\n");
           return (NULL);
 	}
 
@@ -467,7 +493,7 @@ void kissnet_send_rec_packet (int chan, unsigned char *fbuf, int flen)
 	  if (kiss_debug) {
 	    kiss_debug_print (TO_CLIENT, "Fake command prompt", fbuf, flen);
 	  }
-	  strcpy ((char *)kiss_buff, (char *)fbuf);
+	  strlcpy ((char *)kiss_buff, (char *)fbuf, sizeof(kiss_buff));
 	  kiss_len = strlen((char *)kiss_buff);
 	}
 	else {
@@ -485,7 +511,7 @@ void kissnet_send_rec_packet (int chan, unsigned char *fbuf, int flen)
 	    text_color_set(DW_COLOR_DEBUG);
 	    dw_printf ("\n");
 	    dw_printf ("Packet content before adding KISS framing and any escapes:\n");
-	    hex_dump ((char*)fbuf, flen);
+	    hex_dump (fbuf, flen);
 	  }
 
 	  kiss_len = kiss_encapsulate (stemp, flen+1, kiss_buff);
