@@ -14,7 +14,6 @@ void multi_modem_init (struct audio_s *pmodem);
 
 void multi_modem_process_sample (int c, int audio_sample);
 
-void multi_modem_process_rec_frame (int chan, int subchan, unsigned char *fbuf, int flen, int level, retry_t retries);
-
+void multi_modem_process_rec_frame (int chan, int subchan, int slice, unsigned char *fbuf, int flen, alevel_t alevel, retry_t retries);
 
 #endif
