@@ -1,7 +1,7 @@
 //
 //    This file is part of Dire Wolf, an amateur radio packet TNC.
 //
-//    Copyright (C) 2011, 2012, 2013  John Langner, WB2OSZ
+//    Copyright (C) 2011, 2012, 2013, 2014, 2015  John Langner, WB2OSZ
 //
 //    This program is free software: you can redistribute it and/or modify
 //    it under the terms of the GNU General Public License as published by
@@ -23,13 +23,13 @@
  * File:	rrbb.c
  *
  * Purpose:	Raw Received Bit Buffer.
- *		Implementation of an array of bits used to hold data out of
+ *		An array of bits used to hold data out of
  *		the demodulator before feeding it into the HLDC decoding.
  *
- * Version 1.0:	Let's try something new.
- *		Rather than storing a single bit from the demodulator
- *		output, let's store a value which we can try later
- *		comparing to threshold values besides 0.
+ * Version 1.2: Save initial state of 9600 baud descrambler so we can
+ *		attempt bit fix up on G3RUH/K9NG scrambled data.
+ *
+ * Version 1.3:	Store as bytes rather than packing 8 bits per byte.
  *
  *******************************************************************************/
 
@@ -46,46 +46,9 @@
 #include "rrbb.h"
 
 
-
-
 #define MAGIC1 0x12344321
 #define MAGIC2 0x56788765
 
-#ifndef SLICENDICE
-static const unsigned int masks[SOI] = {
-	0x00000001,
-	0x00000002,
-	0x00000004,
-	0x00000008,
-	0x00000010,
-	0x00000020,
-	0x00000040,
-	0x00000080,
-	0x00000100,
-	0x00000200,
-	0x00000400,
-	0x00000800,
-	0x00001000,
-	0x00002000,
-	0x00004000,
-	0x00008000,
-	0x00010000,
-	0x00020000,
-	0x00040000,
-	0x00080000,
-	0x00100000,
-	0x00200000,
-	0x00400000,
-	0x00800000,
-	0x01000000,
-	0x02000000,
-	0x04000000,
-	0x08000000,
-	0x10000000,
-	0x20000000,
-	0x40000000,
-	0x80000000 };
-#endif
 
 static int new_count = 0;
 static int delete_count = 0;
@@ -101,36 +64,44 @@ static int delete_count = 0;
  *
  *		subchan	- Which demodulator of the channel.
  *
+ *		slice	- multiple thresholds per demodulator.
+ *
  *		is_scrambled - Is data scrambled? (true, false)
  *
  *		descram_state - State of data descrambler.
  *
+ *		prev_descram - Previous descrambled bit.
+ *
  * Returns:	Handle to be used by other functions.
  *		
  * Description:	
  *
  ***********************************************************************************/
 
-rrbb_t rrbb_new (int chan, int subchan, int is_scrambled, int descram_state)
+rrbb_t rrbb_new (int chan, int subchan, int slice, int is_scrambled, int descram_state, int prev_descram)
 {
 	rrbb_t result;
 
-	assert (SOI == 8 * sizeof(unsigned int));
-
 	assert (chan >= 0 && chan < MAX_CHANS);
 	assert (subchan >= 0 && subchan < MAX_SUBCHANS);
-
+	assert (slice >= 0 && slice < MAX_SLICERS);
 
 	result = malloc(sizeof(struct rrbb_s));
 
 	result->magic1 = MAGIC1;
 	result->chan = chan;
 	result->subchan = subchan;
+	result->slice = slice;
 	result->magic2 = MAGIC2;
 
 	new_count++;
 
-	rrbb_clear (result, is_scrambled, descram_state);
+	if (new_count > delete_count + 100) {
+	  text_color_set(DW_COLOR_ERROR);
+	  dw_printf ("MEMORY LEAK, rrbb_new, new_count=%d, delete_count=%d\n", new_count, delete_count);
+	}
+
+	rrbb_clear (result, is_scrambled, descram_state, prev_descram);
 
 	return (result);
 }
@@ -141,27 +112,39 @@ rrbb_t rrbb_new (int chan, int subchan, int is_scrambled, int descram_state)
  *
  * Purpose:	Clear by setting length to zero, etc.
  *
- * Inputs:	Handle for sample array.
+ * Inputs:	b 		-Handle for sample array.
+ *
+ *		is_scrambled 	- Is data scrambled? (true, false)
+ *
+ *		descram_state 	- State of data descrambler.
+ *
+ *		prev_descram 	- Previous descrambled bit.
  *
  ***********************************************************************************/
 
-void rrbb_clear (rrbb_t b, int is_scrambled, int descram_state)
+void rrbb_clear (rrbb_t b, int is_scrambled, int descram_state, int prev_descram)
 {
 	assert (b != NULL);
 	assert (b->magic1 == MAGIC1);
 	assert (b->magic2 == MAGIC2);
 
 	assert (is_scrambled == 0 || is_scrambled == 1);
+	assert (prev_descram == 0 || prev_descram == 1);
 
 	b->nextp = NULL;
-	b->audio_level = 9999;
+
+	b->alevel.rec = 9999;	// TODO: was there some reason for this instead of 0 or -1?
+	b->alevel.mark = 9999;
+	b->alevel.space = 9999;
+
 	b->len = 0;
 
 	b->is_scrambled = is_scrambled;
 	b->descram_state = descram_state;
-
+	b->prev_descram = prev_descram;
 }
 
+
 /***********************************************************************************
  *
  * Name:	rrbb_append_bit	
@@ -173,46 +156,8 @@ void rrbb_clear (rrbb_t b, int is_scrambled, int descram_state)
  *
  ***********************************************************************************/
 
-#if SLICENDICE
-void rrbb2_append_bit (rrbb_t b, float val)
-{
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-	assert (b->len >= 0);
+/* Definition in header file so it can be inlined. */
 
-	if (b->len >= MAX_NUM_BITS) {
-	  return;	/* Silently discard if full. */
-	}
-
-	b->data[b->len++] = (int)(val * 1000.);
-}
-#else
-void rrbb_append_bit (rrbb_t b, int val)
-{
-	unsigned int di, mi;
-	
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-
-	if (b->len >= MAX_NUM_BITS) {
-	  return;	/* Silently discard if full. */
-	}
-
-	di = b->len / SOI;
-	mi = b->len % SOI;
-
-	if (val) {
-	  b->data[di] |= masks[mi];
-	}
-	else {
-	  b->data[di] &= ~ masks[mi];
-	}
-
-	b->len++;
-}
-#endif
 
 /***********************************************************************************
  *
@@ -258,92 +203,6 @@ int rrbb_get_len (rrbb_t b)
 }
 
 
-/***********************************************************************************
- *
- * Name:	rrbb_set_slice_val	
- *
- * Purpose:	Set slicing value to determine whether a sample is bit 0 or 1.
- *
- * Inputs:	Handle for sample array.
- *		Slicing point value.
- *		
- ***********************************************************************************/
-
-#if SLICENDICE
-
-static int cmp_slice (slice_t *a, slice_t *b)
-{
-	return ( *a - *b );
-}
-
-void rrbb_set_slice_val (rrbb_t b, slice_t slice_val)
-{
-	slice_t sorted[MAX_NUM_BITS];
-	int n, i;
-	int sum, ave, median, izero;
-
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-
-	b->slice_val = slice_val;
-
-	memcpy (sorted, b->data, b->len * sizeof(slice_t));
-
-	/* Typically takes 14 milliseconds on a reasonable PC. */
-	qsort (sorted, (size_t)(b->len), sizeof(slice_t), cmp_slice);
-
-	text_color_set (DW_COLOR_DEBUG);
-	
-	n = 0;
-	dw_printf ("[%d..%d] ", n, n+9);
-	for (i=n; i<=n+9; i++) dw_printf (" %d", sorted[i]);
-	dw_printf ("\n");
-
-	n = ( b->len / 2 ) - 10;
-	dw_printf ("m[%d..%d] ", n, n+19);
-	for (i=n; i<=n+19; i++) dw_printf (" %d", sorted[i]);
-	dw_printf ("\n");
-
-	n = b->len - 1 - 9;
-	dw_printf ("[%d..%d] ", n, n+9);
-	for (i=n; i<=n+9; i++) dw_printf (" %d", sorted[i]);
-	dw_printf ("\n");
-
-	sum = 0;
-	for (i=0; i<b->len; i++) {
-	  sum += sorted[i];
-	}
-	ave = sum / b->len;
-
-	//b->slice_val = ave;
-	//b->slice_val = sorted[b->len/2];
-	 
-	/* Find first one >= 0. */
-	izero = -1;
-	for (i=0; i<b->len; i++) {
-	  if (sorted[i] >= 0) {
-	    izero = i;
-	    break;
-	  }
-	}
-
-	if (izero >= 0) {
-	  n = izero - 10;
-	  dw_printf ("z[%d..%d] ", n, n+19);
-	  for (i=n; i<=n+19; i++) dw_printf (" %d", sorted[i]);
-	  dw_printf ("\n");
-
-	b->slice_val = sorted[izero-1];
-
-	}
-	
-	
-
-}
-
-#endif
-
 
 /***********************************************************************************
  *
@@ -356,62 +215,9 @@ void rrbb_set_slice_val (rrbb_t b, slice_t slice_val)
  *		
  ***********************************************************************************/
 
-#if SLICENDICE
-int rrbb_get_bit (rrbb_t b, unsigned int ind)
-{
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-	assert (ind >= 0 && ind < b->len);
+/* Definition in header file so it can be inlined. */
 
-	if (b->data[ind] > b->slice_val) {
-	  return 1;
-	}
-	else {
-	  return 0;
-	}
-}
-#else
-int rrbb_get_bit (rrbb_t b, unsigned int ind)
-{
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
 
-	assert (ind < b->len);
-
-	if (b->data[ind / SOI] & masks[ind % SOI]) {
-	  return 1;
-	}
-	else {
-	  return 0;
-	}
-}
-#endif
-unsigned int rrbb_get_computed_bit (rrbb_t b, unsigned int ind)
-{
-	return b->computed_data[ind];
-}
-
-int rrbb_compute_bits (rrbb_t b)
-{
-	unsigned int i,val;
-
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-
-	for (i=0;i<b->len;i++) {
-	  if (b->data[i / SOI] & masks[i % SOI]) {
-	    val = 1;
-	  }
-	  else {
-	    val = 0;
-	  }
-	  b->computed_data[i] = val;
-	}
-	return 0;
-}
 
 
 /***********************************************************************************
@@ -552,22 +358,44 @@ int rrbb_get_subchan (rrbb_t b)
 
 /***********************************************************************************
  *
- * Name:	rrbb_set_audio_level	
+ * Name:	rrbb_get_slice	
  *
- * Purpose:	Set audio level at time the frame was received.
+ * Purpose:	Get slice number from which bit buffer was received.
  *
  * Inputs:	b	Handle for bit array.
- *		a	Audio level.
  *		
  ***********************************************************************************/
 
-void rrbb_set_audio_level (rrbb_t b, int a)
+int rrbb_get_slice (rrbb_t b)
 {
 	assert (b != NULL);
 	assert (b->magic1 == MAGIC1);
 	assert (b->magic2 == MAGIC2);
 
-	b->audio_level = a;
+	assert (b->slice >= 0 && b->slice < MAX_SLICERS);
+
+	return (b->slice);
+}
+
+
+/***********************************************************************************
+ *
+ * Name:	rrbb_set_audio_level	
+ *
+ * Purpose:	Set audio level at time the frame was received.
+ *
+ * Inputs:	b	Handle for bit array.
+ *		alevel	Audio level.
+ *		
+ ***********************************************************************************/
+
+void rrbb_set_audio_level (rrbb_t b, alevel_t alevel)
+{
+	assert (b != NULL);
+	assert (b->magic1 == MAGIC1);
+	assert (b->magic2 == MAGIC2);
+
+	b->alevel = alevel;
 }
 
 
@@ -581,56 +409,16 @@ void rrbb_set_audio_level (rrbb_t b, int a)
  *		
  ***********************************************************************************/
 
-int rrbb_get_audio_level (rrbb_t b)
+alevel_t rrbb_get_audio_level (rrbb_t b)
 {
 	assert (b != NULL);
 	assert (b->magic1 == MAGIC1);
 	assert (b->magic2 == MAGIC2);
 
-	return (b->audio_level);
+	return (b->alevel);
 }
 
 
-/***********************************************************************************
- *
- * Name:	rrbb_set_fix_bits	
- *
- * Purpose:	Set fix bits at time the frame was received.
- *
- * Inputs:	b	Handle for bit array.
- *		a	fix_bits.
- *		
- ***********************************************************************************/
-
-void rrbb_set_fix_bits (rrbb_t b, int a)
-{
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-
-	b->fix_bits = a;
-}
-
-
-/***********************************************************************************
- *
- * Name:	rrbb_get_fix_bits	
- *
- * Purpose:	Get fix bits at time the frame was received.
- *
- * Inputs:	b	Handle for bit array.
- *		
- ***********************************************************************************/
-
-int rrbb_get_fix_bits (rrbb_t b)
-{
-	assert (b != NULL);
-	assert (b->magic1 == MAGIC1);
-	assert (b->magic2 == MAGIC2);
-
-	return (b->fix_bits);
-}
-
 
 /***********************************************************************************
  *
@@ -675,6 +463,26 @@ int rrbb_get_descram_state (rrbb_t b)
 }
 
 
+/***********************************************************************************
+ *
+ * Name:	rrbb_get_prev_descram	
+ *
+ * Purpose:	Get previous descrambled bit before first data bit of frame.
+ *
+ * Inputs:	b	Handle for bit array.
+ *		
+ ***********************************************************************************/
+
+int rrbb_get_prev_descram (rrbb_t b)
+{
+	assert (b != NULL);
+	assert (b->magic1 == MAGIC1);
+	assert (b->magic2 == MAGIC2);
+
+	return (b->prev_descram);
+}
+
+
 /* end rrbb.c */
 
 
