@@ -1,7 +1,7 @@
 //
 //    This file is part of Dire Wolf, an amateur radio packet TNC.
 //
-//    Copyright (C) 2011,2012,2013,2014  John Langner, WB2OSZ
+//    Copyright (C) 2011, 2012, 2013, 2014, 2015  John Langner, WB2OSZ
 //
 //    This program is free software: you can redistribute it and/or modify
 //    it under the terms of the GNU General Public License as published by
@@ -31,8 +31,12 @@
 #include <stdlib.h>	
 #include <string.h>	
 #include <ctype.h>	
+
+#include "direwolf.h"
 #include "textcolor.h"
 #include "symbols.h"
+#include "tt_text.h"
+
 
 //#if __WIN32__
 	char *strcasestr(const char *S, const char *FIND);
@@ -259,6 +263,17 @@ static const struct {
 	/*  ~  94  */	{ "Q4", "TNC Stream Switch" } };
 
 
+// Make sure the array is null terminated.
+static const char *search_locations[] = {
+	(const char *) "symbols-new.txt",
+#ifndef __WIN32__
+	(const char *) "/usr/share/direwolf/symbols-new.txt",
+	(const char *) "/usr/local/share/direwolf/symbols-new.txt",
+#endif
+	(const char *) NULL
+};
+
+
 /*------------------------------------------------------------------
  *
  * Function:	symbols_init
@@ -306,21 +321,33 @@ static int new_sym_len = 0;			/* Number of elements used. */
 
 void symbols_init (void)
 {
-	FILE *fp;
-	struct {
-	  char overlay;
-	  char symbol;
-	  char sp1;
-	  char equal;
-	  char sp2;
-	  char description[150];
-	} stuff;
+	FILE *fp = NULL;
+
+/*
+ * We only care about lines with this format:
+ *
+ *  Column 1 - overlay character of / \ upper case or digit
+ *  Column 2 - symbol in range of ! thru ~
+ *  Column 3 - space
+ *  Column 4 - equal sign
+ *  Column 5 - space
+ *  Column 6 - Start of description.
+ */
+
+#define COL1_OVERLAY 0
+#define COL2_SYMBOL 1
+#define COL3_SP 2
+#define COL4_EQUAL 3
+#define COL5_SP 4
+#define COL6_DESC 5
+
+	char stuff[200];
 	int j;
 
-#define GOOD_LINE(x) ((x.overlay == '/' || x.overlay == '\\' || isupper(x.overlay) || isdigit(x.overlay)) \
-			&& x.symbol >= '!' && x.symbol <= '~' \
-			&& x.sp1 == ' ' && x.equal == '=' && x.sp2 == ' ')
-
+#define GOOD_LINE(x) (strlen(x) > 6 && \
+			(x[COL1_OVERLAY] == '/' || x[COL1_OVERLAY] == '\\' || isupper(x[COL1_OVERLAY]) || isdigit(x[COL1_OVERLAY])) \
+			&& x[COL2_SYMBOL] >= '!' && x[COL2_SYMBOL] <= '~' \
+			&& x[COL3_SP] == ' ' && x[COL4_EQUAL] == '=' && x[COL5_SP] == ' ' && x[COL6_DESC] != ' ')
 
 	if (new_sym_ptr != NULL) {
 	  return;			/* was called already. */
@@ -328,13 +355,13 @@ void symbols_init (void)
 
 // If search strategy changes, be sure to keep decode_tocall in sync.
 
+	fp = NULL;
+	j = 0;
+	do {
+	  if (search_locations[j] == NULL) break;
+	  fp = fopen(search_locations[j++], "r");
+	} while (fp == NULL);
 
-	fp = fopen("symbols-new.txt", "r");
-#ifndef __WIN32__
-	if (fp == NULL) {
-	  fp = fopen("/usr/share/direwolf/symbols-new.txt", "r");
-	}
-#endif
 	if (fp == NULL) {
 
 	  text_color_set(DW_COLOR_ERROR);
@@ -350,7 +377,7 @@ void symbols_init (void)
 /*
  * Count number of interesting lines and allocate storage.
  */
-	while (fgets((char*)(&stuff), sizeof(stuff), fp) != NULL) {
+	while (fgets(stuff, sizeof(stuff), fp) != NULL) {
 	  if (GOOD_LINE(stuff)) {
 	    new_sym_size++;
 	  }
@@ -363,15 +390,15 @@ void symbols_init (void)
  */
 	rewind (fp);
 
-	while (fgets((char*)(&stuff), sizeof(stuff), fp) != NULL) {
+	while (fgets(stuff, sizeof(stuff), fp) != NULL) {
 
 	  if (GOOD_LINE(stuff)) {
-	    for (j = strlen(stuff.description) - 1; j>=0 && stuff.description[j] <= ' '; j--) {
-	      stuff.description[j] = '\0';
+	    for (j = strlen(stuff+COL6_DESC) - 1; j>=0 && stuff[COL6_DESC+j] <= ' '; j--) {
+	      stuff[COL6_DESC+j] = '\0';
 	    }
-	    new_sym_ptr[new_sym_len].overlay = stuff.overlay;
-	    new_sym_ptr[new_sym_len].symbol = stuff.symbol;
-	    strncpy(new_sym_ptr[new_sym_len].description, stuff.description, NEW_SYM_DESC_LEN);
+	    new_sym_ptr[new_sym_len].overlay = stuff[COL1_OVERLAY];
+	    new_sym_ptr[new_sym_len].symbol = stuff[COL2_SYMBOL];
+	    strncpy(new_sym_ptr[new_sym_len].description, stuff+COL6_DESC, NEW_SYM_DESC_LEN);
 	    new_sym_len++;
 	  }
 	}
@@ -389,6 +416,83 @@ void symbols_init (void)
 } /* end symbols_init */
 
 
+/*------------------------------------------------------------------
+ *
+ * Function:	symbols_list
+ *
+ * Purpose:	Print a list of all the symbols.
+ *
+ * Inputs:	none
+ *
+ *------------------------------------------------------------------*/
+
+void symbols_list (void) 
+{
+	int n;
+
+	dw_printf ("\n");
+
+	dw_printf ("\tPRIMARY SYMBOL TABLE\n");
+	dw_printf ("\n");
+	dw_printf ("sym  GPSxy  GPSCnn  APRStt  Icon\n");
+	dw_printf ("---  -----  ------  ------  ----\n");
+	for (n = 1; n < SYMTAB_SIZE; n++) {
+	  dw_printf (" /%c     %s      %02d  AB1%02d   %s\n", n + ' ', primary_symtab[n].xy, n, n, primary_symtab[n].description);
+	}
+
+	dw_printf ("\n");
+	dw_printf ("\tALTERNATE SYMBOL TABLE\n");
+	dw_printf ("\n");
+	dw_printf ("sym  GPSxy  GPSEnn  APRStt  Icon\n");
+	dw_printf ("---  -----  ------  ------  ----\n");
+	for (n = 1; n < SYMTAB_SIZE; n++) {
+	  dw_printf (" \\%c     %s      %02d  AB2%02d   %s\n", n + ' ', alternate_symtab[n].xy, n, n, alternate_symtab[n].description);
+	}
+
+	dw_printf ("\n");
+	dw_printf ("\tNEW SYMBOLS from symbols-new.txt\n");
+	dw_printf ("\n");
+	dw_printf ("sym  GPSxyz  GPSxnn  APRStt   Icon\n");
+	dw_printf ("---  ------  ------  ------   ----\n");
+
+
+	for (n = 0; n < new_sym_len; n++) {
+
+	  int overlay = new_sym_ptr[n].overlay;
+	  int symbol = new_sym_ptr[n].symbol;
+	  char tones[12];
+
+	  symbols_to_tones (overlay, symbol, tones, sizeof(tones));
+
+	  if (overlay == '/') {
+
+	    dw_printf (" %c%c     %s%c     C%02d  %-7s  %s\n", overlay, symbol, 
+								primary_symtab[symbol - ' '].xy, ' ',
+								symbol - ' ', tones,
+								new_sym_ptr[n].description);
+	  }
+	  else if (isupper(overlay) || isdigit(overlay)) {
+
+	    dw_printf (" %c%c     %s%c          %-7s  %s\n", overlay, symbol, 
+								alternate_symtab[symbol - ' '].xy, overlay,
+								tones,
+								new_sym_ptr[n].description);
+	  }
+	  else {
+
+	    dw_printf (" %c%c     %s%c     E%02d  %-7s  %s\n", overlay, symbol, 
+								alternate_symtab[symbol - ' '].xy, ' ', 
+								symbol - ' ', tones,
+								new_sym_ptr[n].description);
+	  }
+	}
+	dw_printf ("\n");
+
+
+} /* end symbols_list */
+
+
+
 /*------------------------------------------------------------------
  *
  * Function:	symbols_from_dest_or_src
@@ -590,19 +694,19 @@ int symbols_into_dest (char symtab, char symbol, char *dest)
 	if (symbol >= '!' && symbol <= '~' && symtab == '/') {
 	  
 	  /* Primary Symbol table. */
-	  sprintf (dest, "GPSC%02d", symbol - ' ');
+	  snprintf (dest, 7, "GPSC%02d", symbol - ' ');
 	  return (0);
 	}
 	else if (symbol >= '!' && symbol <= '~' && symtab == '\\') {
 	  
 	  /* Alternate Symbol table. */
-	  sprintf (dest, "GPSE%02d", symbol - ' ');
+	  snprintf (dest, 7, "GPSE%02d", symbol - ' ');
 	  return (0);
 	}
 	else if (symbol >= '!' && symbol <= '~' && (isupper(symtab) || isdigit(symtab))) {
 
 	  /* Alternate Symbol table with overlay. */
-	  sprintf (dest, "GPS%s%c", alternate_symtab[symbol - ' '].xy, symtab);
+	  snprintf (dest, 7, "GPS%s%c", alternate_symtab[symbol - ' '].xy, symtab);
 	  return (0);
 	}
 
@@ -611,7 +715,7 @@ int symbols_into_dest (char symtab, char symbol, char *dest)
 	dw_printf ("Could not convert symbol \"%c%c\" to GPSxyz destination format.\n",
 			symtab, symbol);
 
-	strcpy (dest, "GPS???");	/* Error. */
+	strlcpy (dest, "GPS???", sizeof(dest));	/* Error. */
 	return (1);
 }
 
@@ -625,16 +729,19 @@ int symbols_into_dest (char symtab, char symbol, char *dest)
  * Inputs:	symtab		/, \, 0-9, A-Z
  *		symbol		any printable character ! to ~ 
  *
+ *		desc_size	Size of description provided by caller
+ *				so we can avoid buffer overflow.
+ *
  * Outputs:	description	Text description.
  *				"--no-symbol--"  if error.
  *
- *
+ *	 
  * Description:	This is used for the monitoring and the 
  *		decode_aprs utility.
  *
  *------------------------------------------------------------------*/
 
-void symbols_get_description (char symtab, char symbol, char *description)
+void symbols_get_description (char symtab, char symbol, char *description, size_t desc_size)
 {
 	char tmp2[2];
 	int j;
@@ -660,7 +767,7 @@ void symbols_get_description (char symtab, char symbol, char *description)
 	  /* We do the latter. */
 
 	  symbol = ' ';
-	  strcpy (description, primary_symtab[symbol-' '].description);
+	  strlcpy (description, primary_symtab[symbol-' '].description, desc_size);
 	  return;
 	}
 
@@ -676,7 +783,7 @@ void symbols_get_description (char symtab, char symbol, char *description)
 
 	for (j=0; j<new_sym_len; j++) {
 	  if (symtab == new_sym_ptr[j].overlay && symbol == new_sym_ptr[j].symbol) {
-	    strcpy (description, new_sym_ptr[j].description);
+	    strlcpy (description, new_sym_ptr[j].description, desc_size);
 	    return;
 	  }
 	}  
@@ -684,15 +791,15 @@ void symbols_get_description (char symtab, char symbol, char *description)
 // Otherwise use the original symbol tables.
 
 	if (symtab == '/') {
-	  strcpy (description, primary_symtab[symbol-' '].description);
+	  strlcpy (description, primary_symtab[symbol-' '].description, desc_size);
 	}
 	else {
-	  strcpy (description, alternate_symtab[symbol-' '].description);
+	  strlcpy (description, alternate_symtab[symbol-' '].description, desc_size);
 	  if (symtab != '\\') {
-	    strcat (description, " w/overlay ");
+	    strlcat (description, " w/overlay ", desc_size);
 	    tmp2[0] = symtab;
 	    tmp2[1] = '\0';
-	    strcat (description, tmp2);
+	    strlcat (description, tmp2, desc_size);
 	  }
 	}
 
@@ -786,6 +893,57 @@ int symbols_code_from_description (char overlay, char *description, char *symtab
 } /* end symbols_code_from_description */
 
 
+
+/*------------------------------------------------------------------
+ *
+ * Function:	symbols_to_tones
+ *
+ * Purpose:	Convert symbol to APRStt tone sequence.
+ *
+ * Inputs:	symtab/overlay
+ *		symbol
+ *		tonessiz	- Amount of space available for result.
+ *
+ * Output:	tones	- string of AB...		
+ *		
+ * Description: 
+ *
+ *		Primary: 	AB1nn		nn = same number as GPSCnn
+ *		Alternate:	AB2nn 		nn = same number as GPSEnn
+ *		with overlay:	AB0nntt   	nn = same as with alternate
+ *						tt = one or two tones from two key method.
+ *
+ *------------------------------------------------------------------*/
+
+void symbols_to_tones (char symtab, char symbol, char *tones, size_t tonessiz)
+{
+
+	if (symtab == '/') {
+
+	  snprintf (tones, tonessiz, "AB1%02d", symbol - ' ');
+	}
+	else if (isupper(symtab) || isdigit(symtab)) {
+
+	  char text[2];
+	  char tt[8];
+
+	  text[0] = symtab;
+	  text[1] = '\0';
+
+	  tt_text_to_two_key (text, 0, tt);
+
+	  snprintf (tones, tonessiz, "AB0%02d%s", symbol - ' ', tt);
+	}
+	else {
+	 
+	  snprintf (tones, tonessiz, "AB2%02d", symbol - ' ');
+	}
+
+}  /* end symbols_to_tones */
+
+
+
+
 #if 0
 
 /* Quick, incomplete, unit test. */
